import os
import csv 
import bz2
import gzip
import re
import urllib2
import socket

from apache_reader import ApacheLogReader

class LocalStats(object):
    """Base class that writes the log file
    """
    def _get_logs(self, logfile, file_urls):
        """Needs to return an iterator. Each entry
        should be a dictionary"""
        if callable(logfile):
            return logfile(file_urls)
        raise NotImplementedError

    def _get_file_obj(self, path, mode='r', compression=None):
        """returns a file object"""
        if compression == 'bz2':
            return bz2.BZ2File(path, mode)
        elif compression == 'gz':
            return gzip.open(path, mode)
        return open(path, mode)
    
    def _build_stats(self, logfile, fileobj, files_url='/packages', 
                     filter=None, compression=None):
        """Builds a stats file
        
        - logfile: path to the original log file, or callable
        - fileobj : a file object or a path to create a file
        - files_url : a filter that define the beginnin of package urls 
        - filter: if given, a callable that receives the 
        current line. if the callable returns True, 
        the line is not included
        """
        if isinstance(fileobj, str):
            fileobj = self._get_file_obj(fileobj, 'w', compression)
            file_created = True
        else:
            file_created = False

        writer = csv.writer(fileobj)
        downloads = {}
        for log in self._get_logs(logfile, files_url):
            if filter is not None:
                if filter(log):
                    continue
            filename = log['filename']
            user_agent = log['useragent'] 
            package_name = log['packagename']
            key = (filename, user_agent, package_name)
            count = log.get('count', 1)
            if key in downloads:
                downloads[key] += count
            else:
                downloads[key] = count
        filenames = downloads.keys()
        filenames.sort()
        for key in filenames:
            filename, user_agent, package_name = key
            count = downloads[key]
            writer.writerow((package_name, filename, user_agent, count))
        if file_created:
            fileobj.close()

    def build_daily_stats(self, year, month, day, logfile, fileobj,
                          files_url='/packages', compression=None):
        """creates a daily stats file using an apache log file.
        
        - year, month, day: values for the day 
        - logfile : path to the log file, or callable
        - fileobj : a file object or a path to create a file
        - files_url : a filter that define the beginning of package urls
        """
        def _filter(log):
            return (day != log['day'] or month != log['month'] or 
                    year != log['year'])

        self._build_stats(logfile, fileobj, files_url, _filter, compression)


    def build_monthly_stats(self, year, month, logfile, fileobj,
                            files_url='/packages', compression=None):
        """creates a monthly stats file using an apache log file.
        
        - year, month: values for the month
        - logfile : path to the log file
        - fileobj : a file object or a path to create a file
        - files_url : a filter that define the beginnin of package urls
        """
        def _filter(log):
            return (month != log['month'] or year != log['year'])

        self._build_stats(logfile, fileobj, files_url, _filter, compression)

    def read_stats(self, stats_file):
        """Returns an iterator over a stats file"""
        if isinstance(stats_file, str):
            ext = os.path.splitext(stats_file)[-1][1:]
            stats_file = self._get_file_obj(stats_file, 'r', ext)
        reader = csv.reader(stats_file)
        for line in reader:
            yield {'packagename': line[0],
                   'filename': line[1],
                   'useragent': line[2],
                   'count': line[3]}
        #reader.close()

    def build_local_stats(self, year, month, day, logfile, directory=None):
        """builds local stats with default values"""
        filename = '%d-%.2d-%.2d.bz2' % (year, month, day)
        if directory is not None:
            filename = os.path.join(directory, filename)

        self.build_daily_stats(year, month, day, logfile, filename, 
                               compression='bz2')

class ApacheLocalStats(LocalStats):
    """concrete class that uses the ApacheLogReader"""
    def _get_logs(self, logfile, files_url):
        return ApacheLogReader(logfile, files_url)

class ApacheDistantLocalStats(ApacheLocalStats):
    """Concrete class that gets the data from a distant file"""
    is_url = re.compile(r'^http://')

    def __init__(self, cache_folder='', timeout=5):
        self.cache_folder = cache_folder
        if not os.path.exists(cache_folder):
            os.makedirs(cache_folder)
        self.timeout = timeout

    def get_and_cache(self, url):
        """retrieve the distant file and add it in the local 
        cache"""
        basename = url.split('/')[-1]
        filename = os.path.join(self.cache_folder, basename)
        if os.path.exists(filename):
            # in cache, let's return it
            return filename, open(filename)
        
        # not in cache, we need to retrieve it
        # and store it
        oldtimeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(self.timeout)
        try:
            try:
                content = urllib2.urlopen(url).read()
            except (urllib2.URLError, socket.timeout):
                return '', None
        finally:
            socket.setdefaulttimeout(oldtimeout)

        f = open(filename, 'w')
        try:
            f.write(content)
        finally:
            f.close()
        
        return filename, open(filename)

    def read_stats(self, stats_file):  
        """retrieve a distant file and works with it"""
        if self.is_url.search(stats_file) is not None: 
            path, fileobj = self.get_and_cache(stats_file)
            if path == '':
                return iter([])
        return ApacheLocalStats.read_stats(self, path)




"""
Reads apache log files
"""
import bz2
import gzip
import re
import os

# list of recognized user agents
SETUPTOOLS_UA = (re.compile((r'^.* setuptools/(?P<version>[0-9]\..*)$')), 'setuptools/%s')
URLLIB_UA = (re.compile(r'^Python-urllib/(?P<version>[23]\.[0-9])$'), 'Python-urllib/%s')
SAFARI_UA = (re.compile(r'^Mozilla.* .* Version/(?P<version>.*) Safari/.*$'), 'Safari/%s')
GOOGLEBOT = (re.compile(r'Googlebot-Mobile/(?P<version>.*);'), 'Googlebot-Mobile/%s')
MSNBOT = (re.compile(r'^msnbot/(?P<version>.*) '), 'msnbot/%s')
FIREFOX_UA = (re.compile(r'^Mozilla.*? Firefox/(?P<version>[23])\..*$'), 'Firefox/%s')
PLAIN_MOZILLA = (re.compile(r'^Mozilla/(?P<version>.*?) '), 'Mozilla/%s')

logre = re.compile(r"\[(?P<day>..)/(?P<month>...)/(?P<year>....):"
                   r"(?P<hour>..):(?P<min>..):(?P<sec>..) "
                   r'(?P<zone>.*)\] "GET (?P<path>[^ "]+) HTTP/1.." 200 .*? (?:".*?")? '
                   r'"(User-Agent: )?(?P<useragent>.*)"$', re.DOTALL)

month_names=['jan','feb','mar','apr','may','jun',
             'jul','aug','sep','oct','nov','dec']
month_index = {}

for i in range(12):
    month_index[month_names[i]] = i+1

def month_to_index(month):
    return month_index[month.lower()]

class ApacheLogReader(object):
    """provides an iterator over apache logs"""

    def __init__(self, filename, files_url='', mode=None):
        if mode is None:
            ext = os.path.splitext(filename)[-1]
            if ext in ('.bz2', '.gz'):
                mode = 'r:%s' % ext[1:]
            else:
                mode = 'r'
        if ':' in mode:
            mode, compr = mode.split(':')
        else:
            mode, compr = mode, None
        if compr not in ('bz2', 'gz', None):
            raise ValueError('%s mode not supported' % compr)
        if compr == 'bz2':
            self._data = bz2.BZ2File(filename, mode)
        elif compr == 'gz':
            self._data = gzip.open(filename)
        else:
            self._data = open(filename, mode)

        self.files_url = files_url

    def __iter__(self):
        return self

    def package_name(self, path):
        path = [p for p in path.split('/') if p != '']
        return path[-2]

    def get_simplified_ua(self, user_agent):
        """returns a simplified version of the user agent""" 
        for expr, repl in (URLLIB_UA, SETUPTOOLS_UA, SAFARI_UA, GOOGLEBOT, 
                           MSNBOT, FIREFOX_UA, PLAIN_MOZILLA):
            res = expr.search(user_agent)
            if res is not None:
                return repl % res.group('version')
        return user_agent

    def next(self):

        while True:
            line = self._data.next().strip() 
            m = logre.search(line)
            if m is None:
                continue
            path = m.group('path')
            filename = os.path.basename(path)
            filename = filename.split('?')[0]
            if not path.startswith(self.files_url) or filename == '':
                continue
            res = m.groupdict()
            res['month'] = month_to_index(res['month'])
            res['useragent'] = self.get_simplified_ua(res['useragent'])
            res['filename'] = filename
            res['packagename'] = self.package_name(path)
            res['day'] = int(res['day'])
            res['year'] = int(res['year'])
            res['hour'] = int(res['hour'])
            res['minute'] = int(res['min'])
            return res

        raise StopIteration




import sqlite3, os

class SqliteStorage(object):

    schema = ["create table if not exists files(project, filename, etag)",
              "create index if not exists files_project on files(project)",
              "create index if not exists files_filename on files(filename)",
              "create table if not exists running(pid)",
              ]

    def __init__(self, filename):
        self.conn = sqlite3.connect(filename)
        cursor = self.conn.cursor()
        for stmt in self.schema:
            cursor.execute(stmt)
        self.commit()

    def commit(self):
        self.conn.commit()

    def files(self, project):
        cursor = self.conn.cursor()
        cursor.execute("select filename from files where project=?", (project,))
        return set(r[0] for r in cursor.fetchall())

    def etag(self, filename):
        cursor = self.conn.cursor()
        cursor.execute("select etag from files where filename=?", (filename,))
        res = cursor.fetchone()
        if res:
            return res[0]
        else:
            return None

    def add_file(self, project, filename, etag):
        cursor = self.conn.cursor()
        cursor.execute("insert into files(project, filename, etag) values(?, ?, ?)",
                       (project, filename, etag))

    def remove_file(self, filename):
        cursor = self.conn.cursor()
        cursor.execute("delete from files where filename=?", (filename,))

    def find_running(self):
        cursor = self.conn.cursor()
        cursor.execute("select pid from running")
        res = cursor.fetchall()
        if not res:
            return None
        return res[0][0]

    def start_running(self, pid):
        cursor = self.conn.cursor()
        cursor.execute("insert into running(pid) values(?)", (pid,))

    def end_running(self):
        cursor = self.conn.cursor()
        cursor.execute("delete from running")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from distutils.core import setup
from pip_mirror import version
from setuptools import setup

setup(name='pip_mirror',
      version=version,
      description='Mirroring tool that implements a local pip mirror via PyPi',
      long_description=open('README.md').read(),
      author='Ian Maguire',
      author_email='mr.scalability@gmail.com',
      license = 'Creative Commons version 4.0',
      url='https://github.com/ianmaguire/pip_mirror',
      packages=['pip_mirror'],
      scripts=['scripts/pip_mirror', 'scripts/processlogs', 'scripts/pip_mirror_check_files'],
      keywords = ['pip', 'mirror', 'pep381', 'bandersnatch'],
      classifiers = [],
     )



from __future__ import with_statement
import cPickle, os, xmlrpclib, time, urllib2, httplib, socket
from xml.etree import ElementTree
from functools import partial
import xml.parsers.expat
import sqlite

# library config
pypi = 'pypi.python.org'
BASE = 'https://'+pypi
SIMPLE = BASE + '/simple/'
version = '0.2'
UA = 'pip_mirror/'+version

# package filter
package_filter_count = [0]
def package_filter(pkg,st=None):
    if st is not None and not st.quiet:
        global package_filter_count
        if package_filter_count[0] == 0:
            print "package_filter: started filtering OSS-only packages"
        if package_filter_count[0] % 500 == 0:
            print "package_filter: processed %d packages" % package_filter_count[0]
        package_filter_count[0] += 1

    try:
        last_release = xmlrpc().package_releases(pkg)[-1]
    except IndexError:
        # if no releases available, ignore this package
        return False
    rel_data = xmlrpc().release_data(pkg,last_release) 

    osi_approved = False
    for c in rel_data['classifiers']:
        if c.startswith('License :: OSI Approved'):
            osi_approved = True

    if not osi_approved:
        if (rel_data['license'] in set(['BSD','GPL','MIT','ZPL 2.1','LGPL','GPLv3',
          'MIT License','BSD License','GPL version 2','ZPL','PSF',
          'Apache License 2.0','Apache 2.0','GPL-3','Apache','MIT license',
          'Apache License, Version 2.0','GPLv3+','gpl','Apache Software License',
          'LGPLv3','Public Domain','Simplified BSD','GPLv2','GPL2','MIT/X11',
          'GPL v3','GNU GPL v3','GNU GPL-3 or later','GNU GPL v2','GNU LGPL',
          'GPL v2','GNU General Public License (GPL)',
          'GNU LESSER GENERAL PUBLIC LICENSE, Version 3','New BSD License','LGPL v3',
          'GPL3','The MIT License (MIT)','GNU Lesser General Public License (LGPL)',
          'GNU LESSER GENERAL PUBLIC LICENSE, Version 2.1','Apache2','GNU GPL, GNU LGPL',
          'WTFPL','LGPL License','Public Domain (WTFPL)','GPL3, see LICENSE.txt',
          'CDDL','BSD 3 Clause','3-clause BSD','AGPLv3','New BSD license',
          'GNU General Public License version 3.0 (GPLv3)','GNU GPLv2',
          'Apache License','GNU GPL v2 or later','Boost Software License','MPL',
          'BSD (3-clause)','http://www.fsf.org/licensing/licenses/lgpl.txt',
          'BSD-style license','GNU LGPL v.3','LGPLv2+','LGPLv3+','GNU Public License v3',
          'http://www.apache.org/licenses/LICENSE-2.0','public domain',
          'GNU General Public License version 3 or later','GPL 2',
          'GNU Lesser General Public License','LGPL-3.0',
          'MIT License, Zope Public License (rest.py), BSD License (odict.py)',
          'License :: OSI Approved :: MIT License','Simplified BSD License',
          'AGPLv3 or Permissive for use with Open Source databases','GPL version 3',
          ])):
            osi_approved = True
    return osi_approved

# Helpers

_proxy = None
def xmlrpc():
    global _proxy
    if _proxy is None:
        _proxy = xmlrpclib.ServerProxy(BASE+'/pypi')
        _proxy.useragent = UA
    return _proxy

_conn = None
def http():
    global _conn
    if _conn is None:
        _conn = httplib.HTTPSConnection(pypi)
        _conn.connect()
    # check that connection is still open
    try:
        if not _conn.sock:
            # HTTPS server had announced to close the connection
            raise socket.error
        _conn.sock.getpeername()
    except socket.error:
        _conn = httplib.HTTPSConnection(pypi)
        _conn.connect()
    return _conn

def now():
    return int(time.time())

# Main class

class Synchronization:
    "Picklable status of a mirror"

    def __init__(self):
        self.homedir = None
        self.quiet = False
        
        # time stamps: seconds since 1970
        self.last_completed = 0 # when did the last run complete
        self.last_started = 0   # when did the current run start

        self.complete_projects = set()
        self.projects_to_do = set()

        self.skip_file_contents = False

    def defaults(self):
        # Fill fields that may not exist in the pickle
        for field, value in (('quiet', False),):
            if not hasattr(self, field):
                setattr(self, field, value)

    def store(self):
        with open(os.path.join(self.homedir, "status"), "wb") as f:
            cPickle.dump(self, f, cPickle.HIGHEST_PROTOCOL)
            self.storage.commit()

    @staticmethod
    def load(homedir, storage=None):
        res = cPickle.load(open(os.path.join(homedir, "status"), "rb"))
        res.storage = storage or sqlite.SqliteStorage(os.path.join(homedir, "files"))
        res.defaults()
        res.homedir = homedir # override pickled value in case it got relocated
        return res

    #################### Synchronization logic ##############################

    @staticmethod
    def initialize(targetdir, storage=None):
        'Create a new empty mirror. This operation should not be interrupted.'
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)
        else:
            assert not os.listdir(targetdir)
        for d in ('simple', 'packages', 'serversig',
                  'local-stats/days'):
            os.makedirs(os.path.join(targetdir, 'web', d))
        status = Synchronization()
        status.homedir = targetdir
        status.last_started = now()
        status.projects_to_do = set(xmlrpc().list_packages())
        status.projects_to_do = set( filter(partial(package_filter,st=status), status.projects_to_do) )
        status.storage = storage or sqlite.SqliteStorage(os.path.join(status.homedir, "files"))
        status.store()
        return status

    def synchronize(self):
        # check whether another job is already running
        pid = self.storage.find_running()
        if pid:
            # check whether process still runs
            try:
                os.kill(pid, 0)
            except OSError:
                # process is gone, take over
                self.storage.end_running()
            else:
                self.storage.commit()
                if not self.quiet:
                    print "Currently already running; mirroring is skipped"
                return
        self.storage.start_running(os.getpid())
        self.storage.commit()
        try:
            self._synchronize()
        finally:
            self.storage.end_running()
            self.storage.commit()

    def _synchronize(self):
        'Run synchronization. Can be interrupted and restarted at any time.'
        if self.last_started == 0:
            # no synchronization in progress. Fetch changelog
            self.last_started = now()
            changes = xmlrpc().changelog(self.last_completed-1)
            if not changes:
                self.update_timestamp(self.last_started)
                return
            for change in changes:
                self.projects_to_do.add(change[0])
            self.projects_to_do = set( filter(partial(package_filter,st=self), self.projects_to_do) )
            self.copy_simple_page('')
            self.store()
        # sort projects to allow for repeatable runs
        for project in sorted(self.projects_to_do):
            if not project:
                # skip empty project names resulting from PyPI-wide changes
                continue
            if not self.quiet:
                print "Synchronizing", project.encode('utf-8')
            data = self.copy_simple_page(project)
            if not data:
                self.delete_project(project)
                self.store()
                continue
            try:
                files = set(self.get_package_files(data))
            except xml.parsers.expat.ExpatError, e:
                # not well-formed, skip for now
                if not self.quiet:
                    print "Page for %s cannot be parsed: %r" % (project, e)
                raise
            for file in files:
                if not self.quiet:
                    print "Copying", file
                self.maybe_copy_file(project, file)
            # files start with /; remove it
            relfiles = set(p[1:] for p in files)
            for file in self.storage.files(project)-relfiles:
                    self.remove_file(file)
            self.complete_projects.add(project)
            self.projects_to_do.remove(project)
            self.store()
        self.update_timestamp(self.last_started)
        self.last_completed = self.last_started
        self.last_started = 0
        self.store()

    def update_timestamp(self, when):
        with open(os.path.join(self.homedir, "web", "last-modified"), "wb") as f:
            f.write(time.strftime("%Y%m%dT%H:%M:%S\n", time.gmtime(when)))

    def copy_simple_page(self, project):
        project = project.encode('utf-8')
        h = http()
        if project:
             h.putrequest('GET', '/simple/'+urllib2.quote(project)+'/')
        else:
             h.putrequest('GET', '/simple/')
        h.putheader('User-Agent', UA)
        h.endheaders()
        r = h.getresponse()
        html = r.read()
        if r.status == 404:
            return None
        if r.status == 301:
            # package not existant anymore, however, similarly-spelled
            # package exists
            return None
        if r.status != 200:
            raise ValueError, "Status %d on %s" % (r.status, project)
        project_simple_dir = os.path.join(self.homedir, 'web', 'simple', project)
        if not os.path.exists(project_simple_dir):
            os.mkdir(project_simple_dir)
        with open(os.path.join(project_simple_dir, 'index.html'), "wb") as f:
            f.write(html)
        h.putrequest('GET', '/serversig/'+urllib2.quote(project)+'/')
        h.putheader('User-Agent', UA)
        h.endheaders()
        r = h.getresponse()
        sig = r.read()
        if r.status != 200:
            if not project:
                # index page is unsigned
                return
            raise ValueError, "Status %d on signature for %s" % (r.status, project)
        with open(os.path.join(self.homedir, "web", "serversig", project), "wb") as f:
            f.write(sig)
        return html

    def get_package_files(self, data):
        x = ElementTree.fromstring(data)
        res = []
        for a in x.findall(".//a"):
            url = a.attrib['href']
            if not url.startswith('../../packages/'):
                continue
            url = url.split('#')[0]
            url = url[len('../..'):]
            res.append(url)
        return res

    def maybe_copy_file(self, project, path):
        h = http()
        if self.skip_file_contents:
            h.putrequest("HEAD", urllib2.quote(path))
        else:
            h.putrequest("GET", urllib2.quote(path))
        h.putheader('User-Agent', UA)
        etag = self.storage.etag(path)
        if etag:
            h.putheader("If-none-match", etag)
        h.endheaders()
        r = h.getresponse()
        if r.status == 304:
            # not modified, discard data
            r.read()
            return
        if path.startswith("/"):
            path = path[1:]
        lpath = os.path.join(self.homedir, "web", path)
        if r.status == 200:
            self.storage.remove_file(path) # readd when done downloading
            data = r.read()
            dirname = os.path.dirname(lpath)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(lpath, "wb") as f:
                f.write(data)
            # XXX may set last-modified timestamp on file
            if "etag" in r.msg:
                self.storage.add_file(project, path, r.msg['etag'])
            self.store()
            return
        if r.status == 404:
            self.remove_file(path)

    def remove_file(self, path):
        self.storage.remove_file(path)
        lpath = os.path.join(self.homedir, "web", path)
        if os.path.exists(lpath):
            os.unlink(lpath)

    def delete_project(self, project):
        for f in self.storage.files(project):
            self.remove_file(f)
        project_simple_dir = os.path.join(self.homedir, "web", "simple", project)
        if os.path.exists(project_simple_dir):
            index_file = os.path.join(project_simple_dir, "index.html")
            if os.path.exists(index_file):
                os.unlink(index_file)
            os.rmdir(project_simple_dir)
        project_serversig_dir = os.path.join(self.homedir, "web", "serversig", project)
        if os.path.exists(project_serversig_dir):
            os.unlink(project_serversig_dir)
        if project in self.projects_to_do:
            self.projects_to_do.remove(project)
        if project in self.complete_projects:
            self.complete_projects.remove(project)
